import 'dart:async';

import 'package:adeptannotations/adeptannotations.dart';
import 'package:adeptgenerator/src/base_generator_for_annotation.dart';
import 'package:adeptgenerator/src/config/clone_key_config.dart';
import 'package:analyzer/dart/constant/value.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/nullability_suffix.dart';
import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart';

class CloneGenerator extends BaseGeneratorForAnnotation<Cloneable, CloneKey> {
  final Cloneable globalAnnotation;
  final bool defaultIgnoreSynthetic = false;
  final bool defaultConsiderEmptyCollectionAsNull = true;
  final bool defaultGenerateExtension = true;

  CloneGenerator(BuilderOptions options)
      : globalAnnotation = Cloneable(
          ignoreSynthetic: options.config['clone_ignore_synthetic'],
          considerEmptyCollectionAsNull:
              options.config['clone_consider_empty_collection_as_null'],
          generateCloneExtension:
              options.config['clone_generate_clone_extension'],
        );

  @override
  FutureOr<String> generateForAnnotatedElement(
      Element element, ConstantReader annotation, BuildStep buildStep) {
    ClassElement classElement =
        getClassElement(element, annotationName: '@Cloneable()');
    String cloneFunction = _generateCloneFunction(classElement, annotation);
    return '''
    $cloneFunction
    ''';
  }

  /// This function manages the logic of generating the Clone boilerplate code
  String _generateCloneFunction(
      ClassElement classElement, ConstantReader annotation) {
    checkDefaultConstructor(classElement, annotationName: '@Cloneable()');
    String classType = classElement.displayName;
    String functionName = '\$${classType}Clone';
    String currentDataParam = 'currentData';
    String newDataParam = 'newData';
    String mergeParam = 'merge';
    String functionParams = '{required $classType $currentDataParam, '
        '$classType? $newDataParam, bool $mergeParam = false}';
    String cloneVar = 'clone';

    StringBuffer cloneData = StringBuffer('final $cloneVar = $classType()');
    StringBuffer mergeData = StringBuffer('if($mergeParam){');
    final cloneableAnnotation =
        _getCloneable(classElement) ?? const Cloneable();

    void fieldAnalyzer(FieldElement field) {
      final cloneKey = _getCloneKey(field) ?? const CloneKeyConfig();
      final ignoreSynthetic = cloneKey.ignoreSynthetic ??
          cloneableAnnotation.ignoreSynthetic ??
          globalAnnotation.ignoreSynthetic ??
          defaultIgnoreSynthetic;
      final considerEmptyCollectionAsNull =
          cloneKey.considerEmptyCollectionAsNull ??
              cloneableAnnotation.considerEmptyCollectionAsNull ??
              globalAnnotation.considerEmptyCollectionAsNull ??
              defaultConsiderEmptyCollectionAsNull;

      if (field.isSynthetic && ignoreSynthetic) return;
      if (field.isFinal || field.isConst || field.isStatic) return;
      if (field.getter == null || field.setter == null) return;
      if (cloneKey.ignore) return;

      final isNullable = field.type.nullabilitySuffix != NullabilitySuffix.none;
      final isCollection = field.type.isDartCoreList ||
          field.type.isDartCoreSet ||
          field.type.isDartCoreMap ||
          field.type.isDartCoreIterable;
      String fieldName = field.displayName;
      String currentData = '$currentDataParam.$fieldName';
      String newData = '$newDataParam?.$fieldName';
      if (considerEmptyCollectionAsNull && isCollection) {
        newData =
            '(($newDataParam?.$fieldName${isNullable ? '?' : ''}.isNotEmpty ?? false) ? $newDataParam!.$fieldName : null)';
      }
      String defaultCloneDataExpression = '$newData ?? $currentData';
      String cloneDataFieldValue = defaultCloneDataExpression;

      //Default value applies only if field is nullable, otherwise is not necessary because of null-safety
      bool defaultValueApplies = isNullable;
      if (cloneKey.forceCurrentData) {
        cloneDataFieldValue = currentData;
      } else if (cloneKey.forceNewData) {
        cloneDataFieldValue = newData;
        //When newData is forced then defaultValue applies because newData could be null
        defaultValueApplies = true;
      } else if (cloneKey.cloneFunction != null) {
        cloneDataFieldValue = cloneKey.cloneFunction!;
      }

      if (cloneKey.defaultValue != null && defaultValueApplies) {
        cloneDataFieldValue =
            '($cloneDataFieldValue) ?? ${cloneKey.defaultValue}';
      } else if (cloneKey.forceNewData && !isNullable) {
        cloneDataFieldValue =
            '$defaultCloneDataExpression // This field has no "defaultValue" and has "forceNewData: true" but the field is not nullable so a non null value must be used in case $newData is null that\'s why ?? $currentData is used as fallback';
      }

      cloneData.writeln('..$fieldName = $cloneDataFieldValue');
      mergeData.writeln('$currentDataParam.$fieldName = $cloneVar.$fieldName;');
    }

    superTypeVisitor(classElement.supertype, fieldAnalyzer);
    classElement.fields.forEach(fieldAnalyzer);
    cloneData.writeln(';');
    mergeData.writeln('}');

    final generateExtension = cloneableAnnotation.generateCloneExtension ??
        globalAnnotation.generateCloneExtension ??
        defaultGenerateExtension;

    return '''
      /// This function creates a new instance of [$classType] by creating a 
      /// copy of [$currentDataParam] and replacing its fields with the data in 
      /// [$newDataParam].
      /// Notes:
      /// 1. The function will use the values from [$newDataParam], in case of 
      /// null then $currentDataParam values will be used. 
      /// 2. If [$mergeParam] is true then [$currentDataParam] will mirror the  
      /// new instance returned.
      /// 3. If @CloneKey(forceNewData: true) annotation is used in some field, 
      /// then [$newDataParam] value will be forced for the specified field 
      /// no matter it's nullability 
      $classType $functionName($functionParams){
        ${cloneData.toString()}
        ${mergeData.toString()}
        return $cloneVar;
      }
      
      ${!generateExtension ? '' : '''
      extension \$${classType}CloneExtension on $classType {
          $classType clone({$classType? data, bool merge = false}) {
          return $functionName(currentData: this, newData: data, merge: merge);
        }
      }
      '''}
    ''';
  }

  /// Gets the Cloneable annotation for a given ClassElement
  Cloneable? _getCloneable(ClassElement classElement) {
    DartObject? cloneableAnnotation = annotationChecker
        .firstAnnotationOf(classElement, throwOnUnresolved: false);

    if (cloneableAnnotation != null) {
      return Cloneable(
        ignoreSynthetic:
            cloneableAnnotation.getField('ignoreSynthetic')?.toBoolValue(),
        generateCloneExtension: cloneableAnnotation
            .getField('generateCloneExtension')
            ?.toBoolValue(),
        considerEmptyCollectionAsNull: cloneableAnnotation
            .getField('considerEmptyCollectionAsNull')
            ?.toBoolValue(),
      );
    }
    return null;
  }

  /// Gets the CloneKey annotation for a given FieldElement
  CloneKeyConfig? _getCloneKey(FieldElement field) {
    Iterable<DartObject> annotations = keyAnnotationChecker
        .annotationsOfExact(field, throwOnUnresolved: false);
    int annotationsCount = annotations.length;
    if (annotationsCount > 1) {
      throw InvalidGenerationSourceError(
          'Repeated `@CloneKey` annotation on ${field.displayName}.',
          todo: 'Check the `@CloneKey` annotations on field '
              '${field.displayName}.',
          element: field);
    }
    if (annotationsCount == 1) {
      DartObject annotation = annotations.first;

      String? cloneFunctionDefinition;
      final function = annotation.getField('cloneFunction')?.toFunctionValue();
      if (function != null) {
        String enclosingClass = function.enclosingElement.displayName;
        String functionName = function.displayName;
        String functionParams =
            'currentData: currentData.${field.displayName}, newData: newData?.${field.displayName}';
        cloneFunctionDefinition =
            '$enclosingClass.$functionName($functionParams)';
      }

      Object? defaultValue = annotationValue(annotation, field, 'defaultValue');
      final key = CloneKeyConfig(
        ignore: annotation.getField('ignore')?.toBoolValue(),
        ignoreSynthetic: annotation.getField('ignoreSynthetic')?.toBoolValue(),
        forceCurrentData:
            annotation.getField('forceCurrentData')?.toBoolValue(),
        forceNewData: annotation.getField('forceNewData')?.toBoolValue(),
        considerEmptyCollectionAsNull:
            annotation.getField('considerEmptyCollectionAsNull')?.toBoolValue(),
        defaultValue: defaultValue,
        cloneFunction: cloneFunctionDefinition,
      );
      if (key.forceCurrentData && key.forceNewData) {
        throw InvalidGenerationSourceError(
            'forceCurrentData and forceNewData properties on '
            '${field.displayName} are mutually exclusive and can\'t be true '
            'on the same field.',
            todo: 'Check the `@CloneKey` properties on field '
                '${field.displayName}.',
            element: field);
      }
      return key;
    }
    return null;
  }
}
